@@ -29,6 +29,7 @@
 #include "pcm_local.h"
 #include "pcm_ioplug.h"
 #include "pcm_ext_parm.h"
+#include "pcm_generic.h"
 
 #ifndef PIC
 /* entry for static linking */
@@ -47,6 +48,7 @@ typedef struct snd_pcm_ioplug_priv {
 } ioplug_priv_t;
 
 /* update the hw pointer */
+/* called in lock */
 static void snd_pcm_ioplug_hw_ptr_update(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
@@ -59,7 +61,7 @@ static void snd_pcm_ioplug_hw_ptr_update(snd_pcm_t *pcm)
 			delta = hw - io->last_hw;
 		else
 			delta = pcm->buffer_size + hw - io->last_hw;
-		io->data->hw_ptr += delta;
+		snd_pcm_mmap_hw_forward(io->data->pcm, delta);
 		io->last_hw = hw;
 	} else
 		io->data->state = SNDRV_PCM_STATE_XRUN;
@@ -87,6 +89,9 @@ static int snd_pcm_ioplug_channel_info(snd_pcm_t *pcm, snd_pcm_channel_info_t *i
 static int snd_pcm_ioplug_status(snd_pcm_t *pcm, snd_pcm_status_t * status)
 {
 	ioplug_priv_t *io = pcm->private_data;
+#ifdef PATCH_IOPLUG_STATUS
+	snd_pcm_ioplug_t *data = io->data;
+#endif
 
 	memset(status, 0, sizeof(*status));
 	snd_pcm_ioplug_hw_ptr_update(pcm);
@@ -94,6 +99,14 @@ static int snd_pcm_ioplug_status(snd_pcm_t *pcm, snd_pcm_status_t * status)
 	status->trigger_tstamp = io->trigger_tstamp;
 	status->avail = snd_pcm_mmap_avail(pcm);
 	status->avail_max = io->avail_max;
+#ifdef PATCH_IOPLUG_STATUS
+	if (data->callback->status) {
+		if (data->callback->status(data) == 0) {
+			status->tstamp = data->status_tstamp;
+			status->delay = data->status_delay;
+		}
+	}
+#endif
 	return 0;
 }
 
@@ -137,15 +150,27 @@ static int snd_pcm_ioplug_reset(snd_pcm_t *pcm)
 static int snd_pcm_ioplug_prepare(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err = 0;
 
-	io->data->state = SND_PCM_STATE_PREPARED;
 	snd_pcm_ioplug_reset(pcm);
-	if (io->data->callback->prepare)
-		return io->data->callback->prepare(io->data);
-	return 0;
+	if (io->data->callback->prepare) {
+		err = snd_pcm_trylock(pcm);
+		if (err != 0) {
+			snd_pcm_unlock(pcm); /* to avoid deadlock */
+			err = io->data->callback->prepare(io->data);
+			snd_pcm_lock(pcm);
+		}else{
+			snd_pcm_unlock(pcm);
+			err = io->data->callback->prepare(io->data);
+		}
+	}
+	if (err < 0)
+		return err;
+	io->data->state = SND_PCM_STATE_PREPARED;
+	return err;
 }
 
-static int hw_params_type[SND_PCM_IOPLUG_HW_PARAMS] = {
+static const int hw_params_type[SND_PCM_IOPLUG_HW_PARAMS] = {
 	[SND_PCM_IOPLUG_HW_ACCESS] = SND_PCM_HW_PARAM_ACCESS,
 	[SND_PCM_IOPLUG_HW_FORMAT] = SND_PCM_HW_PARAM_FORMAT,
 	[SND_PCM_IOPLUG_HW_CHANNELS] = SND_PCM_HW_PARAM_CHANNELS,
@@ -313,7 +338,7 @@ static int snd_pcm_ioplug_hw_refine(snd_pcm_t *pcm, snd_pcm_hw_params_t *params)
 		if (err < 0)
 			return err;
 		change2 |= err;
-		/* periods = buffer_bytes / periods */
+		/* periods = buffer_bytes / period_bytes */
 		err = rule_div(params, SND_PCM_HW_PARAM_PERIODS,
 			       SND_PCM_HW_PARAM_BUFFER_BYTES,
 			       SND_PCM_HW_PARAM_PERIOD_BYTES);
@@ -344,6 +369,26 @@ static int snd_pcm_ioplug_hw_refine(snd_pcm_t *pcm, snd_pcm_hw_params_t *params)
 			return err;
 	}
 
+	/* period_bytes = buffer_bytes / periods */
+	err = rule_div(params, SND_PCM_HW_PARAM_PERIOD_BYTES,
+		       SND_PCM_HW_PARAM_BUFFER_BYTES,
+		       SND_PCM_HW_PARAM_PERIODS);
+	if (err < 0)
+		return err;
+	if (err) {
+		/* update period_size and period_time */
+		change |= err;
+		err = snd_ext_parm_interval_refine(hw_param_interval(params, SND_PCM_HW_PARAM_PERIOD_BYTES),
+						   io->params, SND_PCM_IOPLUG_HW_PERIOD_BYTES);
+		if (err < 0)
+			return err;
+		err = refine_back_time_and_size(params, SND_PCM_HW_PARAM_PERIOD_TIME,
+						SND_PCM_HW_PARAM_PERIOD_SIZE,
+						SND_PCM_HW_PARAM_PERIOD_BYTES);
+		if (err < 0)
+			return err;
+	}
+
 	params->info = SND_PCM_INFO_BLOCK_TRANSFER;
 	p = &io->params[SND_PCM_IOPLUG_HW_ACCESS];
 	if (p->active) {
@@ -408,29 +453,32 @@ static int snd_pcm_ioplug_hw_free(snd_pcm_t *pcm)
 static int snd_pcm_ioplug_sw_params(snd_pcm_t *pcm, snd_pcm_sw_params_t *params)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err;
 
-	if (io->data->callback->sw_params)
-		return io->data->callback->sw_params(io->data, params);
-	return 0;
+	if (!io->data->callback->sw_params)
+		return 0;
+
+	snd_pcm_unlock(pcm); /* to avoid deadlock */
+	err = io->data->callback->sw_params(io->data, params);
+	snd_pcm_lock(pcm);
+
+	return err;
 }
 
 
 static int snd_pcm_ioplug_start(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
-	struct timeval tv;
 	int err;
 	
 	if (io->data->state != SND_PCM_STATE_PREPARED)
-		return -EBUSY;
+		return -EBADFD;
 
 	err = io->data->callback->start(io->data);
 	if (err < 0)
 		return err;
 
-	gettimeofday(&tv, 0);
-	io->trigger_tstamp.tv_sec = tv.tv_sec;
-	io->trigger_tstamp.tv_nsec = tv.tv_usec * 1000L;
+	gettimestamp(&io->trigger_tstamp, pcm->tstamp_type);
 	io->data->state = SND_PCM_STATE_RUNNING;
 
 	return 0;
@@ -439,37 +487,39 @@ static int snd_pcm_ioplug_start(snd_pcm_t *pcm)
 static int snd_pcm_ioplug_drop(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
-	struct timeval tv;
 
 	if (io->data->state == SND_PCM_STATE_OPEN)
 		return -EBADFD;
 
 	io->data->callback->stop(io->data);
 
-	gettimeofday(&tv, 0);
-	io->trigger_tstamp.tv_sec = tv.tv_sec;
-	io->trigger_tstamp.tv_nsec = tv.tv_usec * 1000L;
+	gettimestamp(&io->trigger_tstamp, pcm->tstamp_type);
 	io->data->state = SND_PCM_STATE_SETUP;
 
 	return 0;
 }
 
+/* need own locking */
 static int snd_pcm_ioplug_drain(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err;
 
 	if (io->data->state == SND_PCM_STATE_OPEN)
 		return -EBADFD;
 	if (io->data->callback->drain)
 		io->data->callback->drain(io->data);
-	return snd_pcm_ioplug_drop(pcm);
+	snd_pcm_lock(pcm);
+	err = snd_pcm_ioplug_drop(pcm);
+	snd_pcm_unlock(pcm);
+	return err;
 }
 
 static int snd_pcm_ioplug_pause(snd_pcm_t *pcm, int enable)
 {
 	ioplug_priv_t *io = pcm->private_data;
-	static snd_pcm_state_t states[2] = {
-		SND_PCM_STATE_PAUSED, SND_PCM_STATE_RUNNING
+	static const snd_pcm_state_t states[2] = {
+		SND_PCM_STATE_RUNNING, SND_PCM_STATE_PAUSED
 	};
 	int prev, err;
 
@@ -486,18 +536,29 @@ static int snd_pcm_ioplug_pause(snd_pcm_t *pcm, int enable)
 	return 0;
 }
 
+static snd_pcm_sframes_t snd_pcm_ioplug_rewindable(snd_pcm_t *pcm)
+{
+	return snd_pcm_mmap_hw_rewindable(pcm);
+}
+
 static snd_pcm_sframes_t snd_pcm_ioplug_rewind(snd_pcm_t *pcm, snd_pcm_uframes_t frames)
 {
 	snd_pcm_mmap_appl_backward(pcm, frames);
 	return frames;
 }
 
+static snd_pcm_sframes_t snd_pcm_ioplug_forwardable(snd_pcm_t *pcm)
+{
+	return snd_pcm_mmap_avail(pcm);
+}
+
 static snd_pcm_sframes_t snd_pcm_ioplug_forward(snd_pcm_t *pcm, snd_pcm_uframes_t frames)
 {
 	snd_pcm_mmap_appl_forward(pcm, frames);
 	return frames;
 }
 
+/* need own locking */
 static int snd_pcm_ioplug_resume(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
@@ -507,6 +568,7 @@ static int snd_pcm_ioplug_resume(snd_pcm_t *pcm)
 	return 0;
 }
 
+/* called in lock */
 static snd_pcm_sframes_t ioplug_priv_transfer_areas(snd_pcm_t *pcm,
 						       const snd_pcm_channel_area_t *areas,
 						       snd_pcm_uframes_t offset,
@@ -584,7 +646,7 @@ static snd_pcm_sframes_t snd_pcm_ioplug_mmap_commit(snd_pcm_t *pcm,
 		const snd_pcm_channel_area_t *areas;
 		snd_pcm_uframes_t ofs, frames = size;
 
-		snd_pcm_mmap_begin(pcm, &areas, &ofs, &frames);
+		__snd_pcm_mmap_begin(pcm, &areas, &ofs, &frames);
 		if (ofs != offset)
 			return -EIO;
 		return ioplug_priv_transfer_areas(pcm, areas, offset, frames);
@@ -600,6 +662,8 @@ static snd_pcm_sframes_t snd_pcm_ioplug_avail_update(snd_pcm_t *pcm)
 	snd_pcm_uframes_t avail;
 
 	snd_pcm_ioplug_hw_ptr_update(pcm);
+	if (io->data->state == SND_PCM_STATE_XRUN)
+		return -EPIPE;
 	if (pcm->stream == SND_PCM_STREAM_CAPTURE &&
 	    pcm->access != SND_PCM_ACCESS_RW_INTERLEAVED &&
 	    pcm->access != SND_PCM_ACCESS_RW_NONINTERLEAVED) {
@@ -608,7 +672,7 @@ static snd_pcm_sframes_t snd_pcm_ioplug_avail_update(snd_pcm_t *pcm)
 			snd_pcm_uframes_t offset, size = UINT_MAX;
 			snd_pcm_sframes_t result;
 
-			snd_pcm_mmap_begin(pcm, &areas, &offset, &size);
+			__snd_pcm_mmap_begin(pcm, &areas, &offset, &size);
 			result = io->data->callback->transfer(io->data, areas, offset, size);
 			if (result < 0)
 				return result;
@@ -631,19 +695,40 @@ static int snd_pcm_ioplug_nonblock(snd_pcm_t *pcm, int nonblock)
 static int snd_pcm_ioplug_poll_descriptors_count(snd_pcm_t *pcm)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err = 1;
 
-	if (io->data->callback->poll_descriptors_count)
-		return io->data->callback->poll_descriptors_count(io->data);
-	else
-		return 1;
+	if (io->data->callback->poll_descriptors_count) {
+		err = snd_pcm_trylock(pcm);
+		if (err != 0) {
+			snd_pcm_unlock(pcm); /* to avoid deadlock */
+			err = io->data->callback->poll_descriptors_count(io->data);
+			snd_pcm_lock(pcm);
+		}else{
+			snd_pcm_unlock(pcm);
+			err = io->data->callback->poll_descriptors_count(io->data);
+		}
+	}
+	return err;
 }
 
 static int snd_pcm_ioplug_poll_descriptors(snd_pcm_t *pcm, struct pollfd *pfds, unsigned int space)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err;
 
-	if (io->data->callback->poll_descriptors)
-		return io->data->callback->poll_descriptors(io->data, pfds, space);
+	if (io->data->callback->poll_descriptors) {
+		err = snd_pcm_trylock(pcm);
+		if (err != 0) {
+			snd_pcm_unlock(pcm); /* to avoid deadlock */
+			err = io->data->callback->poll_descriptors(io->data, pfds, space);
+			snd_pcm_lock(pcm);
+			return err;
+		}else{
+			snd_pcm_unlock(pcm);
+			err = io->data->callback->poll_descriptors(io->data, pfds, space);
+			return err;
+		}
+	}
 	if (pcm->poll_fd < 0)
 		return -EIO;
 	if (space >= 1 && pfds) {
@@ -658,12 +743,23 @@ static int snd_pcm_ioplug_poll_descriptors(snd_pcm_t *pcm, struct pollfd *pfds,
 static int snd_pcm_ioplug_poll_revents(snd_pcm_t *pcm, struct pollfd *pfds, unsigned int nfds, unsigned short *revents)
 {
 	ioplug_priv_t *io = pcm->private_data;
+	int err;
 
-	if (io->data->callback->poll_revents)
-		return io->data->callback->poll_revents(io->data, pfds, nfds, revents);
-	else
+	if (io->data->callback->poll_revents) {
+		err = snd_pcm_trylock(pcm);
+		if (err != 0) {
+			snd_pcm_unlock(pcm); /* to avoid deadlock */
+			err = io->data->callback->poll_revents(io->data, pfds, nfds, revents);
+			snd_pcm_lock(pcm);
+		}else{
+			snd_pcm_unlock(pcm);
+			err = io->data->callback->poll_revents(io->data, pfds, nfds, revents);
+		}
+	} else {
 		*revents = pfds->revents;
-	return 0;
+		err = 0;
+	}
+	return err;
 }
 
 static int snd_pcm_ioplug_mmap(snd_pcm_t *pcm ATTRIBUTE_UNUSED)
@@ -683,6 +779,36 @@ static int snd_pcm_ioplug_munmap(snd_pcm_t *pcm ATTRIBUTE_UNUSED)
 	return 0;
 }
 
+static snd_pcm_chmap_query_t **snd_pcm_ioplug_query_chmaps(snd_pcm_t *pcm)
+{
+	ioplug_priv_t *io = pcm->private_data;
+
+	if (io->data->version >= 0x010002 &&
+	    io->data->callback->query_chmaps)
+		return io->data->callback->query_chmaps(io->data);
+	return NULL;
+}
+
+static snd_pcm_chmap_t *snd_pcm_ioplug_get_chmap(snd_pcm_t *pcm)
+{
+	ioplug_priv_t *io = pcm->private_data;
+
+	if (io->data->version >= 0x010002 &&
+	    io->data->callback->get_chmap)
+		return io->data->callback->get_chmap(io->data);
+	return NULL;
+}
+
+static int snd_pcm_ioplug_set_chmap(snd_pcm_t *pcm, const snd_pcm_chmap_t *map)
+{
+	ioplug_priv_t *io = pcm->private_data;
+
+	if (io->data->version >= 0x010002 &&
+	    io->data->callback->set_chmap)
+		return io->data->callback->set_chmap(io->data, map);
+	return -ENXIO;
+}
+
 static void snd_pcm_ioplug_dump(snd_pcm_t *pcm, snd_output_t *out)
 {
 	ioplug_priv_t *io = pcm->private_data;
@@ -720,7 +846,7 @@ static int snd_pcm_ioplug_close(snd_pcm_t *pcm)
 	return 0;
 }
 
-static snd_pcm_ops_t snd_pcm_ioplug_ops = {
+static const snd_pcm_ops_t snd_pcm_ioplug_ops = {
 	.close = snd_pcm_ioplug_close,
 	.nonblock = snd_pcm_ioplug_nonblock,
 	.async = snd_pcm_ioplug_async,
@@ -733,9 +859,12 @@ static snd_pcm_ops_t snd_pcm_ioplug_ops = {
 	.dump = snd_pcm_ioplug_dump,
 	.mmap = snd_pcm_ioplug_mmap,
 	.munmap = snd_pcm_ioplug_munmap,
+	.query_chmaps = snd_pcm_ioplug_query_chmaps,
+	.get_chmap = snd_pcm_ioplug_get_chmap,
+	.set_chmap = snd_pcm_ioplug_set_chmap,
 };
 
-static snd_pcm_fast_ops_t snd_pcm_ioplug_fast_ops = {
+static const snd_pcm_fast_ops_t snd_pcm_ioplug_fast_ops = {
 	.status = snd_pcm_ioplug_status,
 	.prepare = snd_pcm_ioplug_prepare,
 	.reset = snd_pcm_ioplug_reset,
@@ -747,10 +876,12 @@ static snd_pcm_fast_ops_t snd_pcm_ioplug_fast_ops = {
 	.hwsync = snd_pcm_ioplug_hwsync,
 	.delay = snd_pcm_ioplug_delay,
 	.resume = snd_pcm_ioplug_resume,
-	.link_fd = NULL,
 	.link = NULL,
+	.link_slaves = NULL,
 	.unlink = NULL,
+	.rewindable = snd_pcm_ioplug_rewindable,
 	.rewind = snd_pcm_ioplug_rewind,
+	.forwardable = snd_pcm_ioplug_forwardable,
 	.forward = snd_pcm_ioplug_forward,
 	.writei = snd_pcm_ioplug_writei,
 	.writen = snd_pcm_ioplug_writen,
@@ -758,6 +889,7 @@ static snd_pcm_fast_ops_t snd_pcm_ioplug_fast_ops = {
 	.readn = snd_pcm_ioplug_readn,
 	.avail_update = snd_pcm_ioplug_avail_update,
 	.mmap_commit = snd_pcm_ioplug_mmap_commit,
+	.htimestamp = snd_pcm_generic_real_htimestamp,
 	.poll_descriptors_count = snd_pcm_ioplug_poll_descriptors_count,
 	.poll_descriptors = snd_pcm_ioplug_poll_descriptors,
 	.poll_revents = snd_pcm_ioplug_poll_revents,
@@ -846,8 +978,13 @@ callback.
 
 Finally, the dump callback is used to print the status of the plugin.
 
+Note that some callbacks (start, stop, pointer, transfer and pause)
+may be called inside the internal pthread mutex, and they shouldn't
+call the PCM functions again unnecessarily from the callback itself;
+otherwise it may lead to a deadlock.
+
 The hw_params constraints can be defined via either
-#snd_pcm_iplug_set_param_minmax() and #snd_pcm_ioplug_set_param_list()
+#snd_pcm_ioplug_set_param_minmax() and #snd_pcm_ioplug_set_param_list()
 functions after calling #snd_pcm_ioplug_create().
 The former defines the minimal and maximal acceptable values for the
 given hw_params parameter (SND_PCM_IOPLUG_HW_XXX).
@@ -887,7 +1024,8 @@ int snd_pcm_ioplug_create(snd_pcm_ioplug_t *ioplug, const char *name,
 	/* We support 1.0.0 to current */
 	if (ioplug->version < 0x010000 ||
 	    ioplug->version > SND_PCM_IOPLUG_VERSION) {
-		SNDERR("ioplug: Plugin version mismatch\n");
+		SNDERR("ioplug: Plugin version mismatch: 0x%x\n",
+		       ioplug->version);
 		return -ENXIO;
 	}
 
@@ -955,7 +1093,7 @@ void snd_pcm_ioplug_params_reset(snd_pcm_ioplug_t *ioplug)
 int snd_pcm_ioplug_set_param_list(snd_pcm_ioplug_t *ioplug, int type, unsigned int num_list, const unsigned int *list)
 {
 	ioplug_priv_t *io = ioplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_IOPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_IOPLUG_HW_PARAMS) {
 		SNDERR("IOPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -979,7 +1117,7 @@ int snd_pcm_ioplug_set_param_list(snd_pcm_ioplug_t *ioplug, int type, unsigned i
 int snd_pcm_ioplug_set_param_minmax(snd_pcm_ioplug_t *ioplug, int type, unsigned int min, unsigned int max)
 {
 	ioplug_priv_t *io = ioplug->pcm->private_data;
-	if (type < 0 && type >= SND_PCM_IOPLUG_HW_PARAMS) {
+	if (type < 0 || type >= SND_PCM_IOPLUG_HW_PARAMS) {
 		SNDERR("IOPLUG: invalid parameter type %d", type);
 		return -EINVAL;
 	}
@@ -1005,6 +1143,10 @@ int snd_pcm_ioplug_reinit_status(snd_pcm_ioplug_t *ioplug)
 {
 	ioplug->pcm->poll_fd = ioplug->poll_fd;
 	ioplug->pcm->poll_events = ioplug->poll_events;
+	if (ioplug->flags & SND_PCM_IOPLUG_FLAG_MONOTONIC)
+		ioplug->pcm->tstamp_type = SND_PCM_TSTAMP_TYPE_MONOTONIC;
+	else
+		ioplug->pcm->tstamp_type = SND_PCM_TSTAMP_TYPE_GETTIMEOFDAY;
 	ioplug->pcm->mmap_rw = ioplug->mmap_rw;
 	return 0;
 }
@@ -1023,3 +1165,19 @@ const snd_pcm_channel_area_t *snd_pcm_ioplug_mmap_areas(snd_pcm_ioplug_t *ioplug
 		return snd_pcm_mmap_areas(ioplug->pcm);
 	return NULL;
 }
+
+/**
+ * \brief Change the ioplug PCM status
+ * \param ioplug the ioplug handle
+ * \param state the PCM status
+ * \return zero if successful or a negative error code
+ *
+ * Changes the PCM status of the ioplug to the given value.
+ * This function can be used for external plugins to notify the status
+ * change, e.g. XRUN.
+ */
+int snd_pcm_ioplug_set_state(snd_pcm_ioplug_t *ioplug, snd_pcm_state_t state)
+{
+	ioplug->state = state;
+	return 0;
+}
